/*
 * Vertex.java
 * Created in 05/04/2006
 */
package feec.ia725.cg.core;

import java.io.Serializable;

/**
 * @author Lucas Shindi Shimo (lshimo@gmail.com)
 * @author Rafael Henrique Manoel (rafael.manoel@gmail.com)
 *
 */
public class Vertex implements Serializable {
    /** Default Serial Version UID. */
    private static final long serialVersionUID = 1L;

    /** X position. */
    private double x;
    /** Y position. */
    private double y;
    /** Z position. */
    private double z;

    /////////////////////////
    //    CONSTRUCTORS     //
    /////////////////////////
    /**
     * @param x1 X position
     * @param y1 Y position
     * @param z1 Z position
     */
    public Vertex(final double x1, final double y1, final double z1) {
        this.x = x1;
        this.y = y1;
        this.z = z1;
    }

    public Vertex(final Vertex vertex) {
        this.x = vertex.getX();
        this.y = vertex.getY();
        this.z = vertex.getZ();
    }

    /////////////////////////////
    //    BUSINESS METHODS     //
    /////////////////////////////

    /**
     * Produto vetorial entre 2 vetores.
     * @param v1 Primeiro vetor.
     * @param v2 Segundo vetor.
     * @return Retorna o vetor resultado.
     */
    public static float[] crossProduct(float[] v1, float[] v2) {
        float[] result = new float[3];
        
        result[0] = v1[1]*v2[2] - v2[1]*v1[2];
        result[1] = v1[2]*v2[0] - v2[2]*v1[0];
        result[2] = v1[0]*v2[1] - v2[0]*v1[1];
        
        return result;
    }
    
    /**
     * Produto vetorial entre este vetor e um outro passado por parametro
     * @param vertex
     * @return
     */
    public Vertex crossProduct(Vertex other) {
    	double x,y,z;
    	
    	x = this.y * other.z - this.z * other.y;
    	y = this.z * other.x - this.x * other.z;
    	z = this.x * other.y - this.y * other.x;
    	
    	return new Vertex(x,y,z);
    }
    
    /**
     * Produto escalar entre 2 vetores.
     * @param v1 Primeiro vetor.
     * @param v2 Segundo vetor.
     * @return Retorna o valor do produto escalar calculado.
     */
    public static float dotProduct(float[] v1, float[] v2) {
        return ((v1[0]*v2[0]) + (v1[1]*v2[1]) + (v1[2]*v2[2]));
    }
    
    /**
     * Calcula a distncia estre dois pontos no espao
     * @param v1 Ponto de referncia
     * @param v2 Ponto de referncia
     * @return distancia
     */
    public static float distance(float[] v1, float[] v2) {
        return (float) Math.sqrt(Math.pow(v1[0] - v2[0], 2)
                + Math.pow(v1[1] - v2[1], 2) + Math.pow(v1[2] - v2[2], 2));
    }
    /**
     * 
     * @param vector
     * @return
     */
    public static void normalizeVector(float[] vector) {
        float dotProd;
        
        dotProd = dotProduct(vector, vector);
        vector[0] = (float) (vector[0]/Math.sqrt(dotProd));
        vector[1] = (float) (vector[1]/Math.sqrt(dotProd));
        vector[2] = (float) (vector[2]/Math.sqrt(dotProd));

    }

    /**
     * 
     * @param vector
     * @return
     */
    public Vertex normalizeVector() {
        double module = getModule();

        double x = this.x / module;
        double y = this.y / module;
        double z = this.z / module;
        
        return new Vertex(x,y,z);
    }
    
    /**
     * Calcula e retorna o mdulo do vetor
     * @return
     */
    public double getModule() {
    	return Math.sqrt(x*x + y*y + z*z);
    }
    
    /**
     * Mtodo que calcula a diferena entre 2 pontos
     * @param v1 Primeiro ponto.
     * @param v2 Segundo ponto.
     * @return Retorna a diferena
     */
    public static float[] diff(float[] v1, float[] v2) {
        float[] result = new float[3];
        
        result[0] = v1[0] - v2[0];
        result[1] = v1[1] - v2[1];
        result[2] = v1[2] - v2[2];
        
        return result;
    }
    
    /**
     * Soma o vetor posio.
     * @param x1 X Position
     * @param y1 Y Position
     * @param z1 Z Position
     */
    public void sum(final double x1, final double y1, final double z1) {
        this.x += x1;
        this.y += y1;
        this.z += z1;
    }

    /**
     * Soma o vetor posio.
     * @param vertex Vetor posio
     */
    public void sum(final Vertex vertex) {
        this.x += vertex.getX();
        this.y += vertex.getY();
        this.z += vertex.getZ();
    }

    /**
     * Multiplicao por um escalar.
     * @param scalar
     * 	Escalar.
     * @return
     * 	Vetor multiplicado
     */
    public Vertex multiply(final double scalar) {
    	return new Vertex(x * scalar,
    			y * scalar,
    			z * scalar);
    }
    
    public float[] asArray() {
        float vertex[] = new float[] {(float) this.getX(), (float) this.getY(),
                (float) this.getZ()};

        return vertex;
    }

    /**
     * Retorna o ngulo Theta (coordenadas esfricas)
     * @return
     * 	Angulo em graus.
     */
    public double getTheta() {
    	if (this.z == 0) {
    		// retorna o angulo zero 
    		return 0;
    	}

    	// encontrando o ngulo em raianos
    	double theta = Math.atan(Math.sqrt(x*x + y*y) / z);

    	// convertendo para graus e retornando
    	return theta * 180 / Math.PI ;
    }

    /**
     * Retorna o ngulo do vetor projetado no eixo xy
     * @return
     * 	Angulo em graus.
     */
    public double getPhi() {
    	if (this.x == 0) {
    		return y < 0 ? -90 : 90;
    	}
    	
    	// encontrando o ngulo em raianos
    	double phi = Math.atan(this.y / this.x);
    	
//    	System.out.println(phi * 180 / Math.PI);
    	
    	if (x < 0) {
    		phi = Math.PI + phi;
    	}
    	
    	// convertendo para graus e retornando
    	return phi * 180 / Math.PI;    	
    }
    
    /**
     * Retorna o ngulo do vetor projetado no eixo xz
     * @return
     * 	Angulo em graus.
     */
    public double getPsi() {
    	if (this.x == 0) {
    		return 90;
    	}
    	
    	// encontrando o ngulo em raianos
    	double phi = Math.atan(Math.abs(this.z / this.x));
    	
    	// convertendo para graus e retornando
    	return phi * 180 / Math.PI;    	
    }

    
    /////////////////////////////
    //    ATRIBUTES MODIFIES   //
    /////////////////////////////
    public void setVertex(final double x1, final double y1, final double z1) {
        this.x = x1;
        this.y = y1;
        this.z = z1;
    }

    public void setVertex(final Vertex vertex) {
        this.x = vertex.getX();
        this.y = vertex.getY();
        this.z = vertex.getZ();
    }
    /////////////////////////////
    //    GETTERS / SETTERS    //
    /////////////////////////////
    /** @return Returns the x. */
    public final double getX() {
        return this.x;
    }
    /** @param x1 The x to set. */
    public final void setX(final double x1) {
        this.x = x1;
    }
    /** @return Returns the y. */
    public final double getY() {
        return this.y;
    }
    /** @param y1 The y to set. */
    public final void setY(final double y1) {
        this.y = y1;
    }
    /** @return Returns the z. */
    public final double getZ() {
        return this.z;
    }
    /** @param z1 The z to set. */
    public final void setZ(final double z1) {
        this.z = z1;
    }

    //////////////////////////////////////
    //  EQUALS / HASH CODE / TO STRING  //
    //////////////////////////////////////
    /** @see java.lang.Object#equals(java.lang.Object) */
    public boolean equals(final Object obj) {
        if (obj == null) {
            return false;
        }

        if (obj == this) {
            return true;
        }

        if (!(obj instanceof Vertex)) {
            return false;
        }

        Vertex vertex = (Vertex) obj;
        return this.x == vertex.x && this.y == vertex.y && this.z == vertex.z;
    }

    /** @see java.lang.Object#hashCode() */
    public int hashCode() {
        int result = 37;
        result = (int) (result * 37 + this.x);
        result = (int) (result * 37 + this.y);
        result = (int) (result * 37 + this.z);
        return result;
    }

    /** @see java.lang.Object#toString() */
    public String toString() {
        return "(" + this.x + "," + this.y + "," + this.z + ")"; 
    }

    /**
     * Clona o objeto
     */
    public Vertex cloneVertex() {
    	return new Vertex(x, y, z);
    }
}
